import base64
import binascii
from typing import Optional

from google.api_core.exceptions import NotFound
from storages.backends.gcloud import GoogleCloudStorage

from .base import CachingHashStrategy


class GoogleCloudStrategy(CachingHashStrategy[GoogleCloudStorage]):
    delete_not_found_exception = (NotFound,)

    def get_remote_file_hash(self, prefixed_path: str) -> Optional[str]:
        normalized_path = prefixed_path.replace("\\", "/")
        blob = self.remote_storage.bucket.get_blob(normalized_path)
        if blob is None:
            return blob
        md5_base64 = blob._properties["md5Hash"]
        return binascii.hexlify(base64.urlsafe_b64decode(md5_base64)).decode()



from importlib import reload

import pytest
from django.test.utils import override_settings

from collectfasta import settings


@override_settings(FOO=2)
def test_get_setting_returns_valid_value():
    assert 2 == settings._get_setting(int, "FOO", 1)


def test_get_setting_returns_default_value_for_missing_setting():
    assert 1 == settings._get_setting(int, "FOO", 1)


@override_settings(FOO="bar")
def test_get_setting_raises_for_invalid_type():
    with pytest.raises(ValueError):
        settings._get_setting(int, "FOO", 1)


def test_basic_settings():
    with override_settings(
        COLLECTFASTA_DEBUG=True,
        COLLECTFASTA_CACHE="custom",
        COLLECTFASTA_ENABLED=False,
        AWS_IS_GZIPPED=True,
        GZIP_CONTENT_TYPES=("text/css", "text/javascript"),
        COLLECTFASTA_THREADS=0,
    ):
        reload(settings)
        assert settings.debug is True
        assert isinstance(settings.cache_key_prefix, str)
        assert settings.cache == "custom"
        assert settings.enabled is False
        assert isinstance(settings.gzip_content_types, tuple)
        assert settings.threads == 0


def test_settings_with_threads():
    with override_settings(COLLECTFASTA_THREADS=22):
        reload(settings)
        assert settings.threads == 22


@pytest.mark.parametrize(
    "django_settings",
    (
        {"COLLECTFASTA_DEBUG": "True"},
        {"COLLECTFASTA_CACHE_KEY_PREFIX": 1},
        {"COLLECTFASTA_CACHE": None},
        {"COLLECTFASTA_THREADS": None},
        {"COLLECTFASTA_ENABLED": 1},
        {"AWS_IS_GZIPPED": "yes"},
        {"GZIP_CONTENT_TYPES": "not tuple"},
    ),
    ids=lambda x: list(x.keys())[0],
)
def test_invalid_setting_type_raises_value_error(django_settings):
    with override_settings(**django_settings):
        with pytest.raises(ValueError):
            reload(settings)



import functools
import os
import pathlib
import random
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import Callable
from typing import TypeVar
from typing import cast

from django.conf import settings as django_settings
from django.utils.module_loading import import_string
from storages.backends.gcloud import GoogleCloudStorage
from storages.backends.s3boto3 import S3ManifestStaticStorage
from typing_extensions import Final

from collectfasta import settings

static_dir: Final = pathlib.Path(django_settings.STATICFILES_DIRS[0])

F = TypeVar("F", bound=Callable[..., Any])


def make_100_files():
    with ThreadPoolExecutor(max_workers=5) as executor:
        for _ in range(50):
            executor.submit(create_big_referenced_static_file)
        executor.shutdown(wait=True)


def get_fake_client():
    from google.api_core.client_options import ClientOptions
    from google.auth.credentials import AnonymousCredentials
    from google.cloud import storage

    client = storage.Client(
        credentials=AnonymousCredentials(),
        project="test",
        client_options=ClientOptions(api_endpoint=django_settings.GS_CUSTOM_ENDPOINT),
    )
    return client


class GoogleCloudStorageTest(GoogleCloudStorage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if django_settings.GS_CUSTOM_ENDPOINT:
            # Use the fake client if we are using the fake endpoint
            self._client = get_fake_client()


class S3ManifestCustomStaticStorage(S3ManifestStaticStorage):
    location = "prefix"
    manifest_name = "prefixfiles.json"


def create_two_referenced_static_files() -> tuple[pathlib.Path, pathlib.Path]:
    """Create a static file, then another file with a reference to the file"""
    path = create_static_file()
    folder_path = static_dir / (path.stem + "_folder")
    folder_path.mkdir()
    reference_path = folder_path / f"{uuid.uuid4().hex}.html"
    reference_path.write_text(f"{{% static '../{path.name}' %}}")
    return (path, reference_path)


def create_static_file() -> pathlib.Path:
    """Write random characters to a file in the static directory."""
    path = static_dir / f"{uuid.uuid4().hex}.html"
    path.write_text("".join(chr(random.randint(0, 64)) for _ in range(500)))
    return path


def create_big_referenced_static_file() -> tuple[pathlib.Path, pathlib.Path]:
    """Create a big static file, then another file with a reference to the file"""
    path = create_big_static_file()
    reference_path = static_dir / f"{uuid.uuid4().hex}.html"
    reference_path.write_text(f"{{% static '{path.name}' %}}")
    return (path, reference_path)


def create_big_static_file() -> pathlib.Path:
    """Write random characters to a file in the static directory."""
    path = static_dir / f"{uuid.uuid4().hex}.html"
    path.write_text("".join(chr(random.randint(0, 64)) for _ in range(100000)))
    return path


def clean_static_dir() -> None:
    clean_static_dir_recurse(static_dir.as_posix())
    clean_static_dir_recurse(django_settings.AWS_LOCATION)
    clean_static_dir_recurse(S3ManifestCustomStaticStorage.location)


def clean_static_dir_recurse(location: str) -> None:
    try:
        for filename in os.listdir(location):
            file = pathlib.Path(location) / filename
            # don't accidentally wipe the whole drive if someone puts / as location.
            if (
                "collectfasta" in str(file.absolute())
                and ".." not in str(file.as_posix())
                and len(list(filter(lambda x: x == "/", str(file.absolute())))) > 2
            ):
                if file.is_file():
                    file.unlink()
                elif file.is_dir():
                    clean_static_dir_recurse(file.as_posix())
                    file.rmdir()
    except FileNotFoundError:
        pass


def override_setting(name: str, value: Any) -> Callable[[F], F]:
    def decorator(fn: F) -> F:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            original = getattr(settings, name)
            setattr(settings, name, value)
            try:
                return fn(*args, **kwargs)
            finally:
                setattr(settings, name, original)

        return cast(F, wrapper)

    return decorator


def override_storage_attr(name: str, value: Any) -> Callable[[F], F]:
    def decorator(fn: F) -> F:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            storage = import_string(django_settings.STORAGES["staticfiles"]["BACKEND"])
            if hasattr(storage, name):
                original = getattr(storage, name)
            else:
                original = None
            setattr(storage, name, value)
            try:
                return fn(*args, **kwargs)
            finally:
                setattr(storage, name, original)

        return cast(F, wrapper)

    return decorator



from typing import Optional

from django.core.files.storage import FileSystemStorage

from .base import CachingHashStrategy
from .base import HashStrategy


class FileSystemStrategy(HashStrategy[FileSystemStorage]):
    def get_remote_file_hash(self, prefixed_path: str) -> Optional[str]:
        try:
            return self.get_local_file_hash(prefixed_path, self.remote_storage)
        except FileNotFoundError:
            return None


class CachingFileSystemStrategy(
    CachingHashStrategy[FileSystemStorage], FileSystemStrategy
): ...



from unittest import mock

from django.contrib.staticfiles.storage import StaticFilesStorage
from django.test import override_settings as override_django_settings

from collectfasta.management.commands.collectstatic import Command
from collectfasta.tests.utils import clean_static_dir
from collectfasta.tests.utils import create_static_file
from collectfasta.tests.utils import override_setting


class MockPostProcessing(StaticFilesStorage):
    def __init__(self):
        super().__init__()
        self.post_process = mock.MagicMock()


@override_setting("threads", 2)
@override_django_settings(
    STORAGES={
        "staticfiles": {
            "BACKEND": "collectfasta.tests.command.test_post_process.MockPostProcessing",
        },
    },
)
def test_calls_post_process_with_collected_files() -> None:
    clean_static_dir()
    path = create_static_file()

    cmd = Command()
    cmd.run_from_argv(["manage.py", "collectstatic", "--noinput"])
    cmd.storage.post_process.assert_called_once_with(
        {path.name: (mock.ANY, path.name)}, dry_run=False
    )



from io import StringIO
from typing import Any

from django.core.management import call_command


def call_collectstatic(*args: Any, **kwargs: Any) -> str:
    out = StringIO()
    call_command(
        "collectstatic", *args, verbosity=3, interactive=False, stdout=out, **kwargs
    )
    return out.getvalue()



from django.test import override_settings as override_django_settings
from pytest_mock import MockerFixture

from collectfasta.tests.conftest import StrategyFixture
from collectfasta.tests.conftest import live_test
from collectfasta.tests.utils import clean_static_dir
from collectfasta.tests.utils import create_static_file
from collectfasta.tests.utils import override_setting

from .utils import call_collectstatic


@override_django_settings(
    STORAGES={
        "staticfiles": {
            "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
        },
    },
)
def test_disable_collectfasta_with_default_storage() -> None:
    clean_static_dir()
    create_static_file()
    assert "1 static file copied" in call_collectstatic(disable_collectfasta=True)


@live_test
def test_disable_collectfasta(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_static_file()
    assert "1 static file copied" in call_collectstatic(disable_collectfasta=True)


@override_setting("enabled", False)
def test_no_load_with_disable_setting(mocker: MockerFixture) -> None:
    mocked_load_strategy = mocker.patch(
        "collectfasta.management.commands.collectstatic.Command._load_strategy"
    )
    clean_static_dir()
    call_collectstatic()
    mocked_load_strategy.assert_not_called()


def test_no_load_with_disable_flag(mocker: MockerFixture) -> None:
    mocked_load_strategy = mocker.patch(
        "collectfasta.management.commands.collectstatic.Command._load_strategy"
    )
    clean_static_dir()
    call_collectstatic(disable_collectfasta=True)
    mocked_load_strategy.assert_not_called()



import re
import tempfile

from django.contrib.staticfiles.storage import StaticFilesStorage
from django.core.files.storage import FileSystemStorage
from pytest_mock import MockerFixture

from collectfasta.strategies.base import HashStrategy


class Strategy(HashStrategy[FileSystemStorage]):
    def __init__(self) -> None:
        super().__init__(FileSystemStorage())

    def get_remote_file_hash(self, prefixed_path: str) -> None:
        pass


def test_get_file_hash() -> None:
    strategy = Strategy()
    local_storage = StaticFilesStorage()

    with tempfile.NamedTemporaryFile(dir=local_storage.base_location) as f:
        f.write(b"spam")
        hash_ = strategy.get_local_file_hash(f.name, local_storage)
    assert re.fullmatch(r"^[A-z0-9]{32}$", hash_) is not None


def test_should_copy_file(mocker: MockerFixture) -> None:
    strategy = Strategy()
    local_storage = StaticFilesStorage()
    remote_hash = "foo"
    mocker.patch.object(
        strategy, "get_remote_file_hash", mocker.MagicMock(return_value=remote_hash)
    )
    mocker.patch.object(
        strategy, "get_local_file_hash", mocker.MagicMock(return_value=remote_hash)
    )
    assert not strategy.should_copy_file("path", "prefixed_path", local_storage)
    mocker.patch.object(
        strategy, "get_local_file_hash", mocker.MagicMock(return_value="bar")
    )
    assert strategy.should_copy_file("path", "prefixed_path", local_storage)



from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from django.conf import settings as django_settings
from django.contrib.staticfiles.management.commands import collectstatic
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import Storage
from django.core.management.base import CommandParser

from collectfasta import __version__
from collectfasta import settings
from collectfasta.strategies import DisabledStrategy
from collectfasta.strategies import Strategy
from collectfasta.strategies import load_strategy

Task = Tuple[str, str, Storage]


def collect_from_folder(
    storage: Storage, path: str = ""
) -> Generator[tuple[str, str], str, None]:
    folders, files = storage.listdir(path)
    for thefile in files:
        if path:
            prefixed = f"{path}/{thefile}"
        else:
            prefixed = thefile
        yield prefixed, prefixed
    for folder in folders:
        if path:
            folder = f"{path}/{folder}"
        yield from collect_from_folder(storage, folder)


class Command(collectstatic.Command):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.num_copied_files = 0
        self.tasks: List[Task] = []
        self.collectfasta_enabled = settings.enabled
        self.strategy: Strategy = DisabledStrategy(Storage())
        self.found_files: Dict[str, Tuple[Storage, str]] = {}

    @staticmethod
    def _load_strategy() -> Type[Strategy[Storage]]:
        strategy_str = getattr(django_settings, "COLLECTFASTA_STRATEGY", None)
        if strategy_str is not None:
            return load_strategy(strategy_str)

        raise ImproperlyConfigured(
            "No strategy configured, please make sure COLLECTFASTA_STRATEGY is set."
        )

    def get_version(self) -> str:
        return __version__

    def add_arguments(self, parser: CommandParser) -> None:
        super().add_arguments(parser)
        parser.add_argument(
            "--disable-collectfasta",
            action="store_true",
            dest="disable_collectfasta",
            default=False,
            help="Disable Collectfasta.",
        )

    def set_options(self, **options: Any) -> None:
        self.collectfasta_enabled = self.collectfasta_enabled and not options.pop(
            "disable_collectfasta"
        )
        if self.collectfasta_enabled:
            self.strategy = self._load_strategy()(self.storage)
            self.storage = self.strategy.wrap_storage(self.storage)
        super().set_options(**options)

    def second_pass(self, stats: Dict[str, List[str]]) -> Dict[str, List[str]]:
        second_pass_strategy = self.strategy.second_pass_strategy()
        if self.collectfasta_enabled and second_pass_strategy:
            self.copied_files = []
            self.symlinked_files = []
            self.unmodified_files = []
            self.num_copied_files = 0
            source_storage = self.storage
            self.storage = second_pass_strategy.wrap_storage(self.storage)
            self.strategy = second_pass_strategy
            self.log(f"Running second pass with {self.strategy.__class__.__name__}...")
            for f, prefixed in collect_from_folder(source_storage):
                self.maybe_copy_file((f, prefixed, source_storage))
            return {
                "modified": self.copied_files + self.symlinked_files,
                "unmodified": self.unmodified_files,
                "post_processed": self.post_processed_files,
            }

        return stats

    def collect(self) -> Dict[str, List[str]]:
        """
        Override collect to copy files concurrently. The tasks are populated by
        Command.copy_file() which is called by super().collect().
        """
        if not self.collectfasta_enabled or not settings.threads:
            return self.second_pass(super().collect())

        # Store original value of post_process in super_post_process and always
        # set the value to False to prevent the default behavior from
        # interfering when using threads. See maybe_post_process().
        super_post_process = self.post_process
        self.post_process = False

        return_value = super().collect()

        with ThreadPoolExecutor(settings.threads) as pool:
            pool.map(self.maybe_copy_file, self.tasks)

        self.maybe_post_process(super_post_process)
        return_value["post_processed"] = self.post_processed_files
        return self.second_pass(return_value)

    def handle(self, *args: Any, **options: Any) -> Optional[str]:
        """Override handle to suppress summary output."""
        ret = super().handle(**options)
        if not self.collectfasta_enabled:
            return ret
        plural = "" if self.num_copied_files == 1 else "s"
        return f"{self.num_copied_files} static file{plural} copied."

    def maybe_copy_file(self, args: Task) -> None:
        """Determine if file should be copied or not and handle exceptions."""
        path, prefixed_path, source_storage = self.strategy.copy_args_hook(args)
        # Build up found_files to look identical to how it's created in the
        # builtin command's collect() method so that we can run post_process
        # after all parallel uploads finish.
        self.found_files[prefixed_path] = (source_storage, path)

        if self.collectfasta_enabled and not self.dry_run:
            self.strategy.pre_should_copy_hook()

            if not self.strategy.should_copy_file(path, prefixed_path, source_storage):
                self.log(f"Skipping '{path}'")
                self.strategy.on_skip_hook(path, prefixed_path, source_storage)
                return

        self.num_copied_files += 1

        existed = prefixed_path in self.copied_files
        super().copy_file(path, prefixed_path, source_storage)
        copied = not existed and prefixed_path in self.copied_files
        if copied:
            self.strategy.post_copy_hook(path, prefixed_path, source_storage)
        else:
            self.strategy.on_skip_hook(path, prefixed_path, source_storage)

    def copy_file(self, path: str, prefixed_path: str, source_storage: Storage) -> None:
        """
        Append path to task queue if threads are enabled, otherwise copy the
        file with a blocking call.
        """
        args = (path, prefixed_path, source_storage)
        if settings.threads and self.collectfasta_enabled:
            self.tasks.append(args)
        else:
            self.maybe_copy_file(args)

    def delete_file(
        self, path: str, prefixed_path: str, source_storage: Storage
    ) -> bool:
        """Override delete_file to skip modified time and exists lookups."""
        if not self.collectfasta_enabled:
            return super().delete_file(path, prefixed_path, source_storage)

        if self.dry_run:
            self.log(f"Pretending to delete '{path}'")
            return True

        self.log(f"Deleting '{path}' on remote storage")

        try:
            self.storage.delete(prefixed_path)
        except self.strategy.delete_not_found_exception:
            pass

        return True

    def maybe_post_process(self, super_post_process: bool) -> None:
        # This method is extracted and modified from the collect() method of the
        # builtin collectstatic command.
        # https://github.com/django/django/blob/5320ba98f3d253afcaa76b4b388a8982f87d4f1a/django/contrib/staticfiles/management/commands/collectstatic.py#L124

        if not super_post_process or not hasattr(self.storage, "post_process"):
            return

        processor = self.storage.post_process(self.found_files, dry_run=self.dry_run)

        for original_path, processed_path, processed in processor:
            if isinstance(processed, Exception):
                self.stderr.write("Post-processing '%s' failed!" % original_path)
                # Add a blank line before the traceback, otherwise it's
                # too easy to miss the relevant part of the error message.
                self.stderr.write("")
                raise processed
            if processed:
                self.log(
                    f"Post-processed '{original_path}' as '{processed_path}'",
                    level=2,
                )
                self.post_processed_files.append(original_path)
            else:
                self.log("Skipped post-processing '%s'" % original_path)



import string

from django.core.files.storage import FileSystemStorage
from pytest_mock import MockerFixture

from collectfasta import settings
from collectfasta.strategies.base import CachingHashStrategy

hash_characters = string.ascii_letters + string.digits


class Strategy(CachingHashStrategy[FileSystemStorage]):
    def __init__(self) -> None:
        super().__init__(FileSystemStorage())

    def get_remote_file_hash(self, prefixed_path: str) -> None:
        pass


def test_get_cache_key() -> None:
    strategy = Strategy()
    cache_key = strategy.get_cache_key("/some/random/path")
    prefix_len = len(settings.cache_key_prefix)
    # case.assertTrue(cache_key.startswith(settings.cache_key_prefix))
    assert cache_key.startswith(settings.cache_key_prefix)
    assert len(cache_key) == 32 + prefix_len
    expected_chars = hash_characters + "_"
    for c in cache_key:
        assert c in expected_chars


def test_gets_and_invalidates_hash(mocker: MockerFixture) -> None:
    strategy = Strategy()
    expected_hash = "hash"
    mocked = mocker.patch.object(
        strategy,
        "get_remote_file_hash",
        new=mocker.MagicMock(return_value=expected_hash),
    )
    # empty cache
    result_hash = strategy.get_cached_remote_file_hash("path", "prefixed_path")
    assert result_hash == expected_hash
    mocked.assert_called_once_with("prefixed_path")

    # populated cache
    mocked.reset_mock()
    result_hash = strategy.get_cached_remote_file_hash("path", "prefixed_path")
    assert result_hash == expected_hash
    mocked.assert_not_called()

    # test destroy_etag
    mocked.reset_mock()
    strategy.invalidate_cached_hash("path")
    result_hash = strategy.get_cached_remote_file_hash("path", "prefixed_path")
    assert result_hash == expected_hash
    mocked.assert_called_once_with("prefixed_path")


def test_post_copy_hook_primes_cache(mocker: MockerFixture) -> None:
    filename = "123abc"
    expected_hash = "abc123"
    strategy = Strategy()

    mocker.patch.object(
        strategy, "get_local_file_hash", return_value=expected_hash, autospec=True
    )
    strategy.post_copy_hook(filename, filename, strategy.remote_storage)

    assert strategy.get_cached_remote_file_hash(filename, filename) == expected_hash



import os
import shutil

import pytest
from django.conf import settings
from django.test import override_settings as override_django_settings
from pytest_uncollect_if import uncollect_if


def composed(*decs):
    def deco(f):
        for dec in reversed(decs):
            f = dec(f)
        return f

    return deco


S3_STORAGE_BACKEND = "storages.backends.s3.S3Storage"
S3_STATIC_STORAGE_BACKEND = "storages.backends.s3.S3StaticStorage"
S3_MANIFEST_STATIC_STORAGE_BACKEND = "storages.backends.s3.S3ManifestStaticStorage"
S3_CUSTOM_MANIFEST_STATIC_STORAGE_BACKEND = (
    "collectfasta.tests.utils.S3ManifestCustomStaticStorage"
)

GOOGLE_CLOUD_STORAGE_BACKEND = "collectfasta.tests.utils.GoogleCloudStorageTest"
FILE_SYSTEM_STORAGE_BACKEND = "django.core.files.storage.FileSystemStorage"

BOTO3_STRATEGY = "collectfasta.strategies.boto3.Boto3Strategy"
BOTO3_MANIFEST_MEMORY_STRATEGY = (
    "collectfasta.strategies.boto3.Boto3ManifestMemoryStrategy"
)
BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY = (
    "collectfasta.strategies.boto3.Boto3ManifestFileSystemStrategy"
)
GOOGLE_CLOUD_STRATEGY = "collectfasta.strategies.gcloud.GoogleCloudStrategy"
FILE_SYSTEM_STRATEGY = "collectfasta.strategies.filesystem.FileSystemStrategy"
CACHING_FILE_SYSTEM_STRATEGY = (
    "collectfasta.strategies.filesystem.CachingFileSystemStrategy"
)

S3_BACKENDS = [
    S3_STORAGE_BACKEND,
    S3_STATIC_STORAGE_BACKEND,
    S3_MANIFEST_STATIC_STORAGE_BACKEND,
    S3_CUSTOM_MANIFEST_STATIC_STORAGE_BACKEND,
]
BACKENDS = [
    *S3_BACKENDS,
    GOOGLE_CLOUD_STORAGE_BACKEND,
    FILE_SYSTEM_STORAGE_BACKEND,
]

STRATEGIES = [
    BOTO3_STRATEGY,
    BOTO3_MANIFEST_MEMORY_STRATEGY,
    BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY,
    GOOGLE_CLOUD_STRATEGY,
    FILE_SYSTEM_STRATEGY,
    CACHING_FILE_SYSTEM_STRATEGY,
]

COMPATIBLE_STRATEGIES_FOR_BACKENDS = {
    S3_STORAGE_BACKEND: [BOTO3_STRATEGY],
    S3_STATIC_STORAGE_BACKEND: [BOTO3_STRATEGY],
    S3_MANIFEST_STATIC_STORAGE_BACKEND: [
        BOTO3_STRATEGY,
        BOTO3_MANIFEST_MEMORY_STRATEGY,
        BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY,
    ],
    S3_CUSTOM_MANIFEST_STATIC_STORAGE_BACKEND: [
        BOTO3_STRATEGY,
        BOTO3_MANIFEST_MEMORY_STRATEGY,
        BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY,
    ],
    GOOGLE_CLOUD_STORAGE_BACKEND: [GOOGLE_CLOUD_STRATEGY],
    FILE_SYSTEM_STORAGE_BACKEND: [FILE_SYSTEM_STRATEGY, CACHING_FILE_SYSTEM_STRATEGY],
}


def two_n_plus_1(files):
    return files * 2 + 1


def n(files):
    return files


def short_name(backend, strategy):
    return f"{backend.split('.')[-1]}:{strategy.split('.')[-1]}"


def params_for_backends():
    for backend in BACKENDS:
        for strategy in COMPATIBLE_STRATEGIES_FOR_BACKENDS[backend]:
            yield pytest.param(
                (backend, strategy),
                marks=[pytest.mark.backend(backend), pytest.mark.strategy(strategy)],
                id=short_name(backend, strategy),
            )


class StrategyFixture:
    def __init__(self, expected_copied_files, backend, strategy, two_pass):
        self.backend = backend
        self.strategy = strategy
        self.expected_copied_files = expected_copied_files
        self.two_pass = two_pass


@pytest.fixture(params=params_for_backends())
def strategy(request):
    backend, strategy = request.param
    if strategy in (
        BOTO3_MANIFEST_MEMORY_STRATEGY,
        BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY,
    ) and backend in (
        S3_MANIFEST_STATIC_STORAGE_BACKEND,
        S3_CUSTOM_MANIFEST_STATIC_STORAGE_BACKEND,
    ):
        expected_copied_files = two_n_plus_1
    else:
        expected_copied_files = n
    with override_django_settings(
        STORAGES={"staticfiles": {"BACKEND": backend}},
        COLLECTFASTA_STRATEGY=strategy,
    ):
        yield StrategyFixture(
            expected_copied_files,
            backend,
            strategy,
            two_pass=strategy
            in (
                BOTO3_MANIFEST_MEMORY_STRATEGY,
                BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY,
            ),
        )


def uncollect_if_not_s3(strategy: tuple[str, str], **kwargs: dict) -> bool:
    backend, _ = strategy
    return backend not in S3_BACKENDS


def uncollect_if_not_cloud(strategy: tuple[str, str], **kwargs: dict) -> bool:
    backend, _ = strategy
    return backend not in S3_BACKENDS and backend != GOOGLE_CLOUD_STORAGE_BACKEND


live_test = pytest.mark.live_test
speed_test_mark = pytest.mark.speed_test

speed_test = composed(
    live_test,
    speed_test_mark,
    pytest.mark.skipif(
        "not config.getoption('speedtest')",
        reason="no --speedtest flag",
    ),
)

aws_backends_only = uncollect_if(func=uncollect_if_not_s3)
cloud_backends_only = uncollect_if(func=uncollect_if_not_cloud)


def uncollect_if_not_two_pass(strategy: tuple[str, str], **kwargs: dict) -> bool:
    _, strategy_str = strategy
    return strategy_str not in (
        BOTO3_MANIFEST_MEMORY_STRATEGY,
        BOTO3_MANIFEST_FILE_SYSTEM_STRATEGY,
    )


def uncollect_if_two_pass(strategy: tuple[str, str], **kwargs: dict) -> bool:
    return not uncollect_if_not_two_pass(strategy, **kwargs)


two_pass_only = uncollect_if(func=uncollect_if_not_two_pass)
exclude_two_pass = uncollect_if(func=uncollect_if_two_pass)


@pytest.fixture(autouse=True)
def create_test_directories():
    paths = (settings.STATICFILES_DIRS[0], settings.STATIC_ROOT, settings.MEDIA_ROOT)
    for path in paths:
        if not os.path.exists(path):
            os.makedirs(path)
    try:
        yield
    finally:
        for path in paths:
            shutil.rmtree(path)



from typing import Container
from typing import Type
from typing import TypeVar

from django.conf import settings
from typing_extensions import Final

T = TypeVar("T")


def _get_setting(type_: Type[T], key: str, default: T) -> T:
    value = getattr(settings, key, default)
    if not isinstance(value, type_):
        raise ValueError(
            f"The {key!r} setting must be of type {type_!r}, found {type(value)}"
        )
    return value


debug: Final = _get_setting(
    bool, "COLLECTFASTA_DEBUG", _get_setting(bool, "DEBUG", False)
)
cache_key_prefix: Final = _get_setting(
    str, "COLLECTFASTA_CACHE_KEY_PREFIX", "collectfasta06_asset_"
)
cache: Final = _get_setting(str, "COLLECTFASTA_CACHE", "default")
threads: Final = _get_setting(int, "COLLECTFASTA_THREADS", 0)
enabled: Final = _get_setting(bool, "COLLECTFASTA_ENABLED", True)
aws_is_gzipped: Final = _get_setting(bool, "AWS_IS_GZIPPED", False)
gzip_content_types: Final[Container] = _get_setting(
    tuple,
    "GZIP_CONTENT_TYPES",
    (
        "text/css",
        "text/javascript",
        "application/javascript",
        "application/x-javascript",
        "image/svg+xml",
    ),
)



from typing import Any
from typing import Callable
from typing import Optional
from typing import Protocol
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast
from typing import runtime_checkable

from django.contrib.staticfiles.storage import ManifestFilesMixin
from django.core.files.storage import FileSystemStorage
from django.core.files.storage import Storage
from django.core.files.storage.memory import InMemoryStorage
from django.utils.functional import LazyObject

from .base import HashStrategy
from .base import Strategy


@runtime_checkable
class LocationConstructorProtocol(Protocol):

    def __init__(self, location: Optional[str]) -> None: ...


@runtime_checkable
class HasLocationProtocol(Protocol):
    location: str


class InMemoryManifestFilesStorage(ManifestFilesMixin, InMemoryStorage):
    url: Callable[..., Any]


class FileSystemManifestFilesStorage(ManifestFilesMixin, FileSystemStorage):
    url: Callable[..., Any]


OriginalStorage = Union[
    LocationConstructorProtocol,
    HasLocationProtocol,
    Storage,
    ManifestFilesMixin,
    LazyObject,
]


class HashingTwoPassStrategy(HashStrategy[Storage]):
    """
    Hashing strategies interact a lot with the remote storage as a part of post-processing
    This strategy will instead run the hashing strategy using InMemoryStorage first, then copy
    the files to the remote storage
    """

    first_manifest_storage: Type[OriginalStorage]
    second_strategy: Type[Strategy[Storage]]
    original_storage: OriginalStorage
    memory_storage: OriginalStorage

    def __init__(self, remote_storage: OriginalStorage) -> None:
        assert issubclass(self.first_manifest_storage, ManifestFilesMixin)
        assert isinstance(remote_storage, ManifestFilesMixin)
        self.first_pass = True
        self.original_storage = remote_storage
        self.memory_storage = self._get_tmp_storage()
        assert isinstance(self.memory_storage, Storage)
        self.remote_storage = self.memory_storage
        super().__init__(self.memory_storage)

    def _get_tmp_storage(self) -> OriginalStorage:
        # python 3.12 freezes types at runtime, which does not play nicely with
        # LazyObject so we need to cast the type here
        location = cast(HasLocationProtocol, self.original_storage).location
        assert issubclass(self.first_manifest_storage, LocationConstructorProtocol)
        return cast(
            OriginalStorage,
            self.first_manifest_storage(location=location),
        )

    def wrap_storage(self, remote_storage: Storage) -> Storage:
        return self.remote_storage

    def get_remote_file_hash(self, prefixed_path: str) -> Optional[str]:
        try:
            return super().get_local_file_hash(prefixed_path, self.remote_storage)
        except FileNotFoundError:
            return None

    def second_pass_strategy(self):
        """
        Strategy that is used after the first pass of hashing is done - to copy the files
        to the remote destination.
        """
        if self.second_strategy is None:
            raise NotImplementedError(
                "second_strategy must be set to a valid strategy class"
            )
        else:
            assert isinstance(self.original_storage, Storage)
            return self.second_strategy(self.original_storage)


Task = Tuple[str, str, Storage]


class StrategyWithLocationProtocol:
    remote_storage: Any


class WithoutPrefixMixin(StrategyWithLocationProtocol):

    def copy_args_hook(self, args: Task) -> Task:
        assert isinstance(self.remote_storage, HasLocationProtocol)
        if self.remote_storage.location == "" or self.remote_storage.location.endswith(
            "/"
        ):
            location = self.remote_storage.location
        else:
            location = f"{self.remote_storage.location}/"
        return (
            args[0].replace(location, ""),
            args[1].replace(location, ""),
            args[2],
        )


class TwoPassInMemoryStrategy(HashingTwoPassStrategy):
    first_manifest_storage = InMemoryManifestFilesStorage


class TwoPassFileSystemStrategy(HashingTwoPassStrategy):
    first_manifest_storage = FileSystemManifestFilesStorage



import timeit

import pytest
from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings as override_django_settings
from django.test.utils import override_settings
from pytest_mock import MockerFixture

from collectfasta.management.commands.collectstatic import Command
from collectfasta.tests.conftest import StrategyFixture
from collectfasta.tests.conftest import aws_backends_only
from collectfasta.tests.conftest import cloud_backends_only
from collectfasta.tests.conftest import exclude_two_pass
from collectfasta.tests.conftest import live_test
from collectfasta.tests.conftest import speed_test
from collectfasta.tests.conftest import two_pass_only
from collectfasta.tests.utils import clean_static_dir
from collectfasta.tests.utils import create_static_file
from collectfasta.tests.utils import create_two_referenced_static_files
from collectfasta.tests.utils import make_100_files
from collectfasta.tests.utils import override_setting
from collectfasta.tests.utils import override_storage_attr

from .utils import call_collectstatic


@live_test
def test_basics(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_two_referenced_static_files()
    assert (
        f"{strategy.expected_copied_files(2)} static files copied."
        in call_collectstatic()
    )
    # file state should now be cached
    assert "0 static files copied." in call_collectstatic()


@live_test
def test_only_copies_new(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_two_referenced_static_files()
    assert (
        f"{strategy.expected_copied_files(2)} static files copied."
        in call_collectstatic()
    )
    create_two_referenced_static_files()
    # Since the files were already created and are expected to be cached/not copied again,
    # we expect 0 new files to be copied.
    assert (
        f"{strategy.expected_copied_files(2)} static files copied."
        in call_collectstatic()
    )


@live_test
@override_setting("threads", 5)
def test_threads(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_two_referenced_static_files()
    assert (
        f"{strategy.expected_copied_files(2)} static files copied."
        in call_collectstatic()
    )
    # file state should now be cached
    assert "0 static files copied." in call_collectstatic()


@cloud_backends_only
@speed_test
def test_basics_cloud_speed(strategy: StrategyFixture) -> None:
    clean_static_dir()
    make_100_files()
    assert (
        f"{strategy.expected_copied_files(100)} static files copied."
        in call_collectstatic()
    )

    def collectstatic_one():
        assert (
            f"{strategy.expected_copied_files(2)} static files copied."
            in call_collectstatic()
        )

    create_two_referenced_static_files()
    ittook = timeit.timeit(collectstatic_one, number=1)
    print(f"it took {ittook} seconds")


@cloud_backends_only
@speed_test
@override_settings(
    INSTALLED_APPS=["django.contrib.staticfiles"], COLLECTFASTA_STRATEGY=None
)
def test_no_collectfasta_cloud_speed(strategy: StrategyFixture) -> None:
    clean_static_dir()
    make_100_files()
    assert "100 static files copied" in call_collectstatic()

    def collectstatic_one():
        assert "2 static files copied" in call_collectstatic()

    create_two_referenced_static_files()
    ittook = timeit.timeit(collectstatic_one, number=1)
    print(f"it took {ittook} seconds")


@exclude_two_pass
def test_dry_run(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_static_file()
    result = call_collectstatic(dry_run=True)
    assert "1 static file copied." in result
    assert "Pretending to copy" in result
    result = call_collectstatic(dry_run=True)
    assert "1 static file copied." in result
    assert "Pretending to copy" in result
    assert "Pretending to delete" in result


@two_pass_only
def test_dry_run_two_pass(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_static_file()
    result = call_collectstatic(dry_run=True)
    assert "0 static files copied." in result
    assert "Pretending to copy" in result
    result = call_collectstatic(dry_run=True)
    assert "0 static files copied." in result
    assert "Pretending to copy" in result
    assert "Pretending to delete" in result


@aws_backends_only
@live_test
@override_storage_attr("gzip", True)
@override_setting("aws_is_gzipped", True)
def test_aws_is_gzipped(strategy: StrategyFixture) -> None:
    clean_static_dir()
    create_two_referenced_static_files()
    assert (
        f"{strategy.expected_copied_files(2)} static files copied."
        in call_collectstatic()
    )

    # file state should now be cached
    assert "0 static files copied." in call_collectstatic()


@override_django_settings(STORAGES={}, COLLECTFASTA_STRATEGY=None)
def test_raises_for_no_configured_strategy() -> None:
    with pytest.raises(ImproperlyConfigured):
        Command._load_strategy()


@live_test
def test_calls_post_copy_hook(strategy: StrategyFixture, mocker: MockerFixture) -> None:
    post_copy_hook = mocker.patch(
        "collectfasta.strategies.base.Strategy.post_copy_hook", autospec=True
    )
    clean_static_dir()
    (path_one, path_two) = create_two_referenced_static_files()
    cmd = Command()
    cmd.run_from_argv(["manage.py", "collectstatic", "--noinput"])
    post_copy_hook.assert_has_calls(
        [
            mocker.call(mocker.ANY, path_one.name, path_one.name, mocker.ANY),
            mocker.call(
                mocker.ANY,
                f"{path_one.name.replace('.html','')}_folder/{path_two.name}",
                f"{path_one.name.replace('.html','')}_folder/{path_two.name}",
                mocker.ANY,
            ),
        ],
        any_order=True,
    )


@live_test
def test_calls_on_skip_hook(strategy: StrategyFixture, mocker: MockerFixture) -> None:
    on_skip_hook = mocker.patch(
        "collectfasta.strategies.base.Strategy.on_skip_hook", autospec=True
    )
    clean_static_dir()
    (path_one, path_two) = create_two_referenced_static_files()
    cmd = Command()
    cmd.run_from_argv(["manage.py", "collectstatic", "--noinput"])
    on_skip_hook.assert_not_called()
    cmd.run_from_argv(["manage.py", "collectstatic", "--noinput"])
    on_skip_hook.assert_has_calls(
        [
            mocker.call(mocker.ANY, path_one.name, path_one.name, mocker.ANY),
            mocker.call(
                mocker.ANY,
                f"{path_one.name.replace('.html','')}_folder/{path_two.name}",
                f"{path_one.name.replace('.html','')}_folder/{path_two.name}",
                mocker.ANY,
            ),
        ],
        any_order=True,
    )



import abc
import gzip
import hashlib
import logging
import mimetypes
import pydoc
from functools import lru_cache
from io import BytesIO
from typing import ClassVar
from typing import Generic
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import Storage
from django.utils.encoding import force_bytes

from collectfasta import settings

_RemoteStorage = TypeVar("_RemoteStorage", bound=Storage)


cache = caches[settings.cache]
logger = logging.getLogger(__name__)


class Strategy(abc.ABC, Generic[_RemoteStorage]):
    # Exceptions raised by storage backend for delete calls to non-existing
    # objects. The command silently catches these.
    delete_not_found_exception: ClassVar[Tuple[Type[Exception], ...]] = ()

    def __init__(self, remote_storage: _RemoteStorage) -> None:
        self.remote_storage = remote_storage

    def wrap_storage(self, remote_storage: _RemoteStorage) -> _RemoteStorage:
        """
        Wrap the remote storage.
        Allows you to change the remote storage behavior
        just for collectstatic.
        """
        return remote_storage

    @abc.abstractmethod
    def should_copy_file(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> bool:
        """
        Called for each file before copying happens, this method decides
        whether a file should be copied or not. Return False to indicate that
        the file is already up-to-date and should not be copied, or True to
        indicate that it is stale and needs updating.
        """
        ...

    def pre_should_copy_hook(self) -> None:
        """Hook called before calling should_copy_file."""
        ...

    def post_copy_hook(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> None:
        """Hook called after a file is copied."""
        ...

    def on_skip_hook(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> None:
        """Hook called when a file copy is skipped."""
        ...

    def second_pass_strategy(self) -> "Optional[Strategy[_RemoteStorage]]":
        """
        Strategy that is used after the first pass of hashing is done - to copy the files
        to the remote destination.
        """
        return None

    def copy_args_hook(
        self, args: Tuple[str, str, Storage]
    ) -> Tuple[str, str, Storage]:
        """Hook called before copying a file. Use this to modify the path or storage."""
        return args


class HashStrategy(Strategy[_RemoteStorage], abc.ABC):
    use_gzip = False

    def should_copy_file(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> bool:
        local_hash = self.get_local_file_hash(path, local_storage)
        remote_hash = self.get_remote_file_hash(prefixed_path)
        return local_hash != remote_hash

    def get_gzipped_local_file_hash(
        self, uncompressed_file_hash: str, path: str, contents: str
    ) -> str:
        buffer = BytesIO()
        zf = gzip.GzipFile(mode="wb", fileobj=buffer, mtime=0.0)
        zf.write(force_bytes(contents))
        zf.close()
        return hashlib.md5(buffer.getvalue()).hexdigest()

    @lru_cache(maxsize=None)
    def get_local_file_hash(self, path: str, local_storage: Storage) -> str:
        """Create md5 hash from file contents."""
        # Read file contents and handle file closing
        file = local_storage.open(path)
        try:
            contents = file.read()
        finally:
            file.close()

        file_hash = hashlib.md5(contents).hexdigest()

        # Check if content should be gzipped and hash gzipped content
        content_type = mimetypes.guess_type(path)[0] or "application/octet-stream"
        if self.use_gzip and content_type in settings.gzip_content_types:
            file_hash = self.get_gzipped_local_file_hash(file_hash, path, contents)

        return file_hash

    @abc.abstractmethod
    def get_remote_file_hash(self, prefixed_path: str) -> Optional[str]: ...


class CachingHashStrategy(HashStrategy[_RemoteStorage], abc.ABC):
    @lru_cache(maxsize=None)
    def get_cache_key(self, path: str) -> str:
        path_hash = hashlib.md5(path.encode()).hexdigest()
        return settings.cache_key_prefix + path_hash

    def invalidate_cached_hash(self, path: str) -> None:
        cache.delete(self.get_cache_key(path))

    def should_copy_file(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> bool:
        local_hash = self.get_local_file_hash(path, local_storage)
        remote_hash = self.get_cached_remote_file_hash(path, prefixed_path)
        if local_hash != remote_hash:
            # invalidate cached hash, since we expect its corresponding file to
            # be overwritten
            self.invalidate_cached_hash(path)
            return True
        return False

    def get_cached_remote_file_hash(self, path: str, prefixed_path: str) -> str:
        """Cache the hash of the remote storage file."""
        cache_key = self.get_cache_key(path)
        hash_ = cache.get(cache_key, False)
        if hash_ is False:
            hash_ = self.get_remote_file_hash(prefixed_path)
            cache.set(cache_key, hash_)
        return str(hash_)

    def get_gzipped_local_file_hash(
        self, uncompressed_file_hash: str, path: str, contents: str
    ) -> str:
        """Cache the hash of the gzipped local file."""
        cache_key = self.get_cache_key("gzip_hash_%s" % uncompressed_file_hash)
        file_hash = cache.get(cache_key, False)
        if file_hash is False:
            file_hash = super().get_gzipped_local_file_hash(
                uncompressed_file_hash, path, contents
            )
            cache.set(cache_key, file_hash)
        return str(file_hash)

    def post_copy_hook(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> None:
        """Cache the hash of the just copied local file."""
        super().post_copy_hook(path, prefixed_path, local_storage)
        key = self.get_cache_key(path)
        value = self.get_local_file_hash(path, local_storage)
        cache.set(key, value)


class DisabledStrategy(Strategy):
    def should_copy_file(
        self, path: str, prefixed_path: str, local_storage: Storage
    ) -> NoReturn:
        raise NotImplementedError

    def pre_should_copy_hook(self) -> NoReturn:
        raise NotImplementedError


def load_strategy(klass: Union[str, type, object]) -> Type[Strategy[Storage]]:
    if isinstance(klass, str):
        klass = pydoc.locate(klass)
    if not isinstance(klass, type) or not issubclass(klass, Strategy):
        raise ImproperlyConfigured(
            "Configured strategies must be subclasses of %s.%s"
            % (Strategy.__module__, Strategy.__qualname__)
        )
    return klass



import logging
import os
import tempfile
from typing import Any
from typing import Dict
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast

import botocore.exceptions
from boto3.s3.transfer import TransferConfig
from django.contrib.staticfiles.storage import ManifestFilesMixin
from django.core.files.storage import Storage
from django.utils.timezone import make_naive
from storages.backends.s3boto3 import S3Boto3Storage
from storages.backends.s3boto3 import S3ManifestStaticStorage
from storages.backends.s3boto3 import S3StaticStorage
from storages.utils import clean_name
from storages.utils import is_seekable
from storages.utils import safe_join
from storages.utils import setting

from collectfasta import settings

from .base import CachingHashStrategy
from .hashing import TwoPassFileSystemStrategy
from .hashing import TwoPassInMemoryStrategy
from .hashing import WithoutPrefixMixin

logger = logging.getLogger(__name__)


class S3StorageWrapperBase(S3Boto3Storage):
    def __init__(self, *args: Any, original: Any, **kwargs: Any) -> None:
        default_settings = original.get_default_settings()
        for name in default_settings.keys():
            setattr(self, name, default_settings[name])
            if hasattr(original, name):
                setattr(self, name, getattr(original, name))
        for arg in [
            "_bucket",
            "_connections",
            "access_key",
            "secret_key",
            "security_token",
            "config",
            "_transfer_config",
        ]:
            if hasattr(original, arg):
                setattr(self, arg, getattr(original, arg))
        if not hasattr(self, "_transfer_config"):
            # not sure why, but the original doesn't have this attribute
            self._transfer_config = TransferConfig(use_threads=self.use_threads)

        self.preload_metadata = True
        self._entries: Dict[str, str] = {}

    # restores the "preload_metadata" method that was removed in django-storages 1.10
    def _save(self, name, content):
        cleaned_name = clean_name(name)
        name = self._normalize_name(cleaned_name)
        params = self._get_write_parameters(name, content)

        if is_seekable(content):
            content.seek(0, os.SEEK_SET)
        if (
            self.gzip
            and params["ContentType"] in self.gzip_content_types
            and "ContentEncoding" not in params
        ):
            content = self._compress_content(content)
            params["ContentEncoding"] = "gzip"

        obj = self.bucket.Object(name)
        if self.preload_metadata:
            self._entries[name] = obj
        obj.upload_fileobj(content, ExtraArgs=params, Config=self._transfer_config)
        return cleaned_name

    @property
    def entries(self):
        if self.preload_metadata and not self._entries:
            self._entries = {
                entry.key: entry
                for entry in self.bucket.objects.filter(Prefix=self.location)
            }
        return self._entries

    def delete(self, name):
        super().delete(name)
        if name in self._entries:
            del self._entries[name]

    def exists(self, name):
        cleaned_name = self._normalize_name(clean_name(name))
        if self.entries:
            return cleaned_name in self.entries
        return super().exists(name)

    def size(self, name):
        cleaned_name = self._normalize_name(clean_name(name))
        if self.entries:
            entry = self.entries.get(cleaned_name)
            if entry:
                return entry.size if hasattr(entry, "size") else entry.content_length
            return 0
        return super().size(name)

    def get_modified_time(self, name):
        """
        Returns an (aware) datetime object containing the last modified time if
        USE_TZ is True, otherwise returns a naive datetime in the local timezone.
        """
        name = self._normalize_name(clean_name(name))
        entry = self.entries.get(name)
        if entry is None:
            entry = self.bucket.Object(name)
        if setting("USE_TZ"):
            # boto3 returns TZ aware timestamps
            return entry.last_modified
        else:
            return make_naive(entry.last_modified)


class ManifestFilesWrapper(ManifestFilesMixin):
    def __init__(self, *args: Any, original: Any, **kwargs: Any) -> None:
        super().__init__(*args, original=original, **kwargs)
        if original.manifest_storage == original:
            self.manifest_storage = cast(Storage, self)
        else:
            self.manifest_storage = original.manifest_storage
        for arg in [
            "hashed_files",
            "manifest_hash",
            "support_js_module_import_aggregation",
            "patterns",
            "_patterns",
            "hashed_files",
        ]:
            if hasattr(original, arg):
                setattr(self, arg, getattr(original, arg))


class S3StorageWrapper(S3StorageWrapperBase, S3Boto3Storage):
    pass


class S3StaticStorageWrapper(S3StorageWrapperBase, S3StaticStorage):
    pass


class S3ManifestStaticStorageWrapper(
    ManifestFilesWrapper,
    S3StorageWrapperBase,
    S3ManifestStaticStorage,
):
    def _save(self, name, content):
        content.seek(0)
        with tempfile.SpooledTemporaryFile() as tmp:
            tmp.write(content.read())
            return super()._save(name, tmp)


S3Storage = TypeVar(
    "S3Storage", bound=Union[S3Boto3Storage, S3StaticStorage, S3ManifestStaticStorage]
)

S3StorageWrapped = Union[
    S3StaticStorageWrapper, S3ManifestStaticStorageWrapper, S3StorageWrapper
]


class Boto3Strategy(CachingHashStrategy[S3Storage]):
    def __init__(self, remote_storage: S3Storage) -> None:
        self.remote_storage = self.wrapped_storage(remote_storage)
        super().__init__(self.remote_storage)
        self.use_gzip = settings.aws_is_gzipped

    def wrapped_storage(self, remote_storage: S3Storage) -> S3StorageWrapped:
        if isinstance(remote_storage, S3ManifestStaticStorage):
            return S3ManifestStaticStorageWrapper(original=remote_storage)
        elif isinstance(remote_storage, S3StaticStorage):
            return S3StaticStorageWrapper(original=remote_storage)
        elif isinstance(remote_storage, S3Boto3Storage):
            return S3StorageWrapper(original=remote_storage)
        return remote_storage

    def wrap_storage(self, remote_storage: S3Storage) -> S3StorageWrapped:
        return self.remote_storage

    def _normalize_path(self, prefixed_path: str) -> str:
        path = str(safe_join(self.remote_storage.location, prefixed_path))
        return path.replace("\\", "")

    @staticmethod
    def _clean_hash(quoted_hash: Optional[str]) -> Optional[str]:
        """boto returns hashes wrapped in quotes that need to be stripped."""
        if quoted_hash is None:
            return None
        assert quoted_hash[0] == quoted_hash[-1] == '"'
        return quoted_hash[1:-1]

    def get_remote_file_hash(self, prefixed_path: str) -> Optional[str]:
        normalized_path = self._normalize_path(prefixed_path)
        logger.debug("Getting file hash", extra={"normalized_path": normalized_path})
        try:
            hash_: str
            if normalized_path in self.remote_storage.entries:
                hash_ = self.remote_storage.entries[normalized_path].e_tag
            else:
                hash_ = self.remote_storage.bucket.Object(normalized_path).e_tag
        except botocore.exceptions.ClientError:
            logger.debug("Error on remote hash request", exc_info=True)
            return None
        return self._clean_hash(hash_)

    def pre_should_copy_hook(self) -> None:
        if settings.threads:
            logger.info("Resetting connection")
            self.remote_storage._connection = None


class Boto3WithoutPrefixStrategy(WithoutPrefixMixin, Boto3Strategy):
    pass


class Boto3ManifestMemoryStrategy(TwoPassInMemoryStrategy):
    second_strategy = Boto3WithoutPrefixStrategy


class Boto3ManifestFileSystemStrategy(TwoPassFileSystemStrategy):
    second_strategy = Boto3WithoutPrefixStrategy



import base64
import os
import pathlib
import sys
import tempfile

# import python and django versions
from django import get_version
from google.cloud import storage
from google.oauth2 import service_account

base_path = pathlib.Path.cwd()

# Set USE_TZ to True to work around bug in django-storages
USE_TZ = True

SECRET_KEY = "nonsense"
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "test-collectfasta",
    }
}
TEMPLATE_LOADERS = (
    "django.template.loaders.filesystem.Loader",
    "django.template.loaders.app_directories.Loader",
    "django.template.loaders.eggs.Loader",
)
TEMPLATE_DIRS = [str(base_path / "collectfasta/templates")]
INSTALLED_APPS = ("collectfasta", "django.contrib.staticfiles")
STATIC_URL = "/staticfiles/"
# python then django version
AWS_LOCATION = sys.version.split(" ")[0] + "-" + get_version()
GS_LOCATION = sys.version.split(" ")[0] + "-" + get_version()
STATIC_ROOT = str(base_path / "static_root")
MEDIA_ROOT = str(base_path / "fs_remote")
STATICFILES_DIRS = [str(base_path / "static")]
STORAGES = {
    "staticfiles": {
        "BACKEND": "storages.backends.s3.S3Storage",
    },
}
COLLECTFASTA_STRATEGY = "collectfasta.strategies.boto3.Boto3Strategy"
COLLECTFASTA_DEBUG = True

GZIP_CONTENT_TYPES = ("text/plain",)
AWS_IS_GZIPPED = False
AWS_QUERYSTRING_AUTH = False
AWS_DEFAULT_ACL: None = None
S3_USE_SIGV4 = True
AWS_S3_SIGNATURE_VERSION = "s3v4"

# AWS
AWS_STORAGE_BUCKET_NAME = "collectfasta"

FAKE_AWS_ACCESS_KEY_ID = "AKIAIOSFODNN7EXAMPLE"
FAKE_AWS_SECRET_ACCESS_KEY = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
AWS_ACCESS_KEY_ID = os.environ.get("AWS_ACCESS_KEY_ID", default=FAKE_AWS_ACCESS_KEY_ID)
AWS_SECRET_ACCESS_KEY = os.environ.get(
    "AWS_SECRET_ACCESS_KEY",
    default=FAKE_AWS_SECRET_ACCESS_KEY,
)
AWS_S3_REGION_NAME = "ap-southeast-2"
if AWS_ACCESS_KEY_ID == FAKE_AWS_ACCESS_KEY_ID:
    AWS_ENDPOINT_URL = "http://localhost.localstack.cloud:4567"
    AWS_S3_ENDPOINT_URL = "http://localhost.localstack.cloud:4567"

GCLOUD_API_CREDENTIALS_BASE64 = os.environ.get(
    "GCLOUD_API_CREDENTIALS_BASE64", default=None
)
# Google Cloud
if GCLOUD_API_CREDENTIALS_BASE64:
    # live test
    GS_CUSTOM_ENDPOINT = None
    with tempfile.NamedTemporaryFile() as file:
        gcloud_credentials_json = base64.b64decode(GCLOUD_API_CREDENTIALS_BASE64)
        file.write(gcloud_credentials_json)
        file.read()
        GS_CREDENTIALS = service_account.Credentials.from_service_account_file(
            file.name
        )
else:
    GS_CUSTOM_ENDPOINT = "http://127.0.0.1:5050"
    try:
        storage.Client(
            client_options={"api_endpoint": GS_CUSTOM_ENDPOINT},
            use_auth_w_custom_endpoint=False,
        ).create_bucket("collectfasta")
    except Exception:
        pass
GS_BUCKET_NAME = "collectfasta"
